import { ReadonlyState, State } from "./state.js";
import { unreachable } from "./util.js";
export class Member extends State {
    parent;
    #readonly;
    constructor({ parent, value }) {
        super(value);
        this.parent = parent;
        this.listen(() => parent.notify({ kind: ListEventKind.MemberUpdate, list: parent, member: this }));
    }
    getCurrentIndex() {
        return this.parent.array.indexOf(this);
    }
    remove() {
        this.parent.removeMember(this);
    }
    replace(value) {
        this.parent.replaceMember(this, value);
    }
    readonly() {
        // This avoids creating a new ReadonlyMember if it already exists for this member
        return (this.#readonly ??= new ReadonlyMember(this, this.parent.readonly()));
    }
}
export class ReadonlyMember extends ReadonlyState {
    parent;
    constructor(member, parent) {
        super(member.value, member);
        this.parent = parent ?? member.parent.readonly();
    }
}
export var ListEventKind;
(function (ListEventKind) {
    ListEventKind["Append"] = "append";
    ListEventKind["Prepend"] = "prepend";
    ListEventKind["Insert"] = "insert";
    ListEventKind["Remove"] = "remove";
    ListEventKind["Replace"] = "replace";
    ListEventKind["Swap"] = "swap";
    ListEventKind["Move"] = "move";
    ListEventKind["Update"] = "update";
    ListEventKind["MemberUpdate"] = "member-update";
})(ListEventKind || (ListEventKind = {}));
const internal = {
    assertIndexLow(index) {
        if (index < 0)
            throw new RangeError("Index must be non-negative");
    },
    assertIndexHigh(index, list) {
        if (index >= list.size.value)
            throw new RangeError("Index is out of bounds");
    },
    assertIndexInBounds(index, list) {
        this.assertIndexLow(index);
        this.assertIndexHigh(index, list);
    },
    removeMember(list, member) {
        const index = list.array.indexOf(member);
        list.array.splice(index, 1);
        list.size.set(list.size.value - 1);
        return index;
    },
    replaceMember(list, old, value) {
        const index = list.array.indexOf(old);
        list.array[index] = new Member({ parent: list, value });
        return index;
    },
    appendMember(list, member) {
        list.array.push(member);
        list.size.set(list.size.value + 1);
        return list.size.value;
    },
    append(list, value) {
        const member = new Member({ parent: list, value });
        internal.appendMember(list, member);
        return member;
    },
    prependMember(list, member) {
        list.array.unshift(member);
        list.size.set(list.size.value + 1);
        return member;
    },
    prepend(list, value) {
        const member = new Member({ parent: list, value });
        return internal.prependMember(list, member);
    },
    drop(list) {
        const member = list.array.pop();
        if (member)
            list.size.set(list.size.value - 1);
        return member;
    },
    dropFirst(list) {
        const member = list.array.shift();
        if (member)
            list.size.set(list.size.value - 1);
        return member;
    },
    insertMemberBefore(list, member, newMember) {
        const index = list.array.indexOf(member);
        list.array.splice(index, 0, newMember);
        list.size.set(list.size.value + 1);
        return index;
    },
    insertMemberAt(list, index, newMember) {
        internal.assertIndexLow(index);
        // TODO: handle the case where the index is out of bounds
        // Consider whether sparse lists should be supported
        if (index > list.size.value)
            throw new RangeError("Index is out of bounds");
        // Special case: inserting at the end of the list
        if (index === list.size.value)
            return internal.appendMember(list, newMember);
        const member = list.at(index);
        if (!member)
            throw new RangeError(`Could not find member at index ${index}. This is a bug.`);
        return internal.insertMemberBefore(list, member, newMember);
    },
    insertAt(list, index, value) {
        const newMember = new Member({ parent: list, value });
        internal.insertMemberAt(list, index, newMember);
        return newMember;
    },
    removeAt(list, index) {
        internal.assertIndexInBounds(index, list);
        const member = list.at(index);
        if (!member)
            throw new RangeError(`Could not find member at index ${index}. This is a bug.`);
        internal.removeMember(list, member);
        return member;
    },
    replaceAt(list, index, value) {
        internal.assertIndexInBounds(index, list);
        const member = list.at(index);
        if (!member)
            throw new RangeError(`Could not find member at index ${index}. This is a bug.`);
        const newIndex = internal.replaceMember(list, member, value);
        return [member, list.array[newIndex]];
    },
    swap(list, a, b) {
        const aIndex = list.array.indexOf(a);
        const bIndex = list.array.indexOf(b);
        list.array[aIndex] = b;
        list.array[bIndex] = a;
        return [aIndex, bIndex];
    },
};
export class List {
    /** @internal */
    array = [];
    size = new State(0);
    /** @internal */
    listeners = [];
    set(iterable) {
        this.array = [];
        let index = 0;
        for (const value of iterable) {
            const member = new Member({ parent: this, value });
            this.array.push(member);
            index++;
        }
        this.size.set(index);
        this.notify({ kind: ListEventKind.Update, list: this });
    }
    constructor(init = []) {
        // TODO: perhaps this could be lazy?
        this.set(init);
    }
    get first() {
        return this.array[0];
    }
    get last() {
        return this.array[this.size.value - 1];
    }
    static isList(value) {
        return value instanceof List || value instanceof ReadonlyList;
    }
    static isReadonly(value) {
        return value instanceof ReadonlyList;
    }
    *[Symbol.iterator]() {
        for (let i = 0; i < this.size.value; i++)
            yield this.array[i];
    }
    // /** @internal */
    // updateIndices(): number {
    // 	let index = 0;
    // 	for (const member of this) member.index.set(index++);
    // 	return this.size.set(index);
    // }
    removeMember(member) {
        const index = internal.removeMember(this, member);
        this.notify({ kind: ListEventKind.Remove, list: this, member, index });
        return index;
    }
    replaceMember(member, value) {
        const index = internal.replaceMember(this, member, value);
        this.notify({ kind: ListEventKind.Replace, list: this, member, old: member, index });
        return index;
    }
    append(value) {
        const member = internal.append(this, value);
        this.notify({ kind: ListEventKind.Append, member, list: this });
        return member;
    }
    prepend(value) {
        const member = internal.prepend(this, value);
        this.notify({ kind: ListEventKind.Prepend, member, list: this });
        return member;
    }
    dropLast() {
        const last = internal.drop(this);
        // TODO: verify that the index is correct. The list will have shrunk by one,
        // so I'm using the size (which will be last index + 1, equal to previous index)
        if (last)
            this.notify({ kind: ListEventKind.Remove, list: this, member: last, index: this.size.value });
        return last;
    }
    dropFirst() {
        const first = internal.dropFirst(this);
        if (first)
            this.notify({ kind: ListEventKind.Remove, list: this, member: first, index: 0 });
        return first;
    }
    /** @escapes This function escapes the list and returns a plain array. The result is not reactive. */
    toArray() {
        const array = [];
        for (const member of this)
            array.push(member.value);
        return array;
    }
    find(value) {
        for (const member of this)
            if (member.value === value)
                return member;
        return undefined;
    }
    at(index) {
        if (index < 0)
            index = this.size.value + index + 1;
        return this.array[index];
    }
    insertAt(index, value) {
        const newMember = internal.insertAt(this, index, value);
        this.notify({ kind: ListEventKind.Insert, list: this, member: newMember, index });
        return newMember;
    }
    removeAt(index) {
        const member = internal.removeAt(this, index);
        this.notify({ kind: ListEventKind.Remove, list: this, member, index });
        return member;
    }
    replaceAt(index, value) {
        const [old, member] = internal.replaceAt(this, index, value);
        this.notify({ kind: ListEventKind.Replace, list: this, member, old, index });
        return member;
    }
    swap(a, b) {
        const [from, to] = internal.swap(this, a, b);
        this.notify({ kind: ListEventKind.Swap, list: this, a, b, from, to });
    }
    swapBetween(i, j) {
        internal.assertIndexInBounds(i, this);
        internal.assertIndexInBounds(j, this);
        const a = this.at(i);
        const b = this.at(j);
        if (a && b)
            this.swap(a, b);
    }
    moveBefore(before, member) {
        if (member === before)
            return;
        const from = internal.removeMember(this, member);
        const to = internal.insertMemberBefore(this, before, member);
        this.notify({ kind: ListEventKind.Move, list: this, member, from, to });
    }
    moveTo(index, member) {
        internal.assertIndexInBounds(index, this);
        // Special case: moving to the end of the list
        if (index === this.size.value - 1) {
            const from = internal.removeMember(this, member);
            internal.append(this, member.value);
            this.notify({ kind: ListEventKind.Move, list: this, member, from, to: index });
        }
        else {
            const before = this.at(index);
            if (before)
                return this.moveBefore(before, member);
        }
    }
    readonly() {
        return new ReadonlyList(this);
    }
    listen(listener) {
        this.listeners.push(listener);
        return () => {
            this.listeners = this.listeners.filter(l => l !== listener);
        };
    }
    /** @internal */
    notify(event) {
        for (const listener of this.listeners)
            listener(event);
    }
    join(joiner) {
        let init = "";
        for (const member of this)
            init += member.value + joiner;
        const state = new State(init);
        this.listen(() => {
            let update = "";
            for (const member of this)
                update += member.value + joiner;
            state.set(update);
        });
        return state;
    }
    // TODO: consider whether slice should react to source list changes
    /** The sliced list will not react to changes in the original list. */
    slice(start = 0, end = Infinity) {
        const sliced = new List();
        for (const member of this.array.slice(start, end))
            sliced.append(member.value);
        return sliced.readonly();
    }
    reverse() {
        const reversed = new List();
        for (const member of this)
            reversed.prepend(member.value);
        this.listen(update => {
            switch (update.kind) {
                case ListEventKind.Append:
                    return reversed.prepend(update.member.value);
                case ListEventKind.Prepend:
                    return reversed.append(update.member.value);
                case ListEventKind.Insert:
                    return reversed.insertAt(this.size.value - update.index - 1, update.member.value);
                case ListEventKind.Remove:
                    return reversed.removeAt(reversed.size.value - update.index - 1);
                case ListEventKind.Replace:
                    return reversed.replaceAt(reversed.size.value - update.index - 1, update.member.value);
                case ListEventKind.Swap:
                    return reversed.swapBetween(reversed.size.value - update.from - 1, reversed.size.value - update.to - 1);
                case ListEventKind.Move:
                    const member = reversed.at(reversed.size.value - update.from - 1);
                    if (!member)
                        return;
                    return reversed.moveTo(reversed.size.value - update.to - 1, member);
                case ListEventKind.Update:
                    reversed.set([]); // Clear the list first
                    for (const member of this)
                        reversed.prepend(member.value);
                    return;
                case ListEventKind.MemberUpdate:
                    return reversed.at(update.member.getCurrentIndex())?.set(update.member.value);
                default:
                    unreachable(update);
            }
        });
        return reversed.readonly();
    }
    /** Like `Array#map`, but function receives List Members instead of values. Returns a new ReadonlyList. */
    each(modifier) {
        const target = new List();
        for (const member of this)
            target.append(modifier(member));
        this.listen(update => {
            switch (update.kind) {
                case ListEventKind.Append:
                    return target.append(modifier(update.member));
                case ListEventKind.Prepend:
                    return target.prepend(modifier(update.member));
                case ListEventKind.Insert:
                    return target.insertAt(update.index, modifier(update.member));
                case ListEventKind.Remove:
                    return target.removeAt(update.index);
                case ListEventKind.Remove:
                    return target.removeAt(update.index);
                case ListEventKind.Replace:
                    return target.replaceAt(update.index, modifier(update.member));
                case ListEventKind.Swap:
                    return target.swapBetween(update.from, update.to);
                case ListEventKind.Move:
                    const member = target.at(update.from);
                    if (!member)
                        return;
                    member.set(modifier(update.member));
                    return target.moveTo(update.to, member);
                case ListEventKind.Update:
                    return target.set(update.list.array.map(modifier));
                case ListEventKind.MemberUpdate:
                    return target.at(update.member.getCurrentIndex())?.set(modifier(update.member));
                default:
                    unreachable(update);
            }
        });
        return target.readonly();
    }
    sort(compareFn) {
        const sorted = new List();
        sorted.set(this.toArray().sort(compareFn));
        // TODO: implement smarter reactivity when source list is sorted
        this.listen(update => sorted.set(update.list.toArray().sort(compareFn)));
        return sorted;
    }
    filter(filterFn) {
        const filtered = new List();
        let index = 0;
        for (const member of this) {
            if (filterFn(member.value, index))
                filtered.append(member.value);
            index++;
        }
        // TODO: we should somehow translate the events from the original list to the filtered list
        // This involves mapping the index of the event to the index of the filtered list
        this.listen(update => filtered.set(update.list.toArray().filter(filterFn)));
        return filtered;
    }
}
export class ReadonlyList {
    /** @internal */
    source;
    size;
    listen;
    at;
    reverse;
    slice;
    toArray;
    join;
    constructor(source) {
        this.source = source;
        this.size = source.size.readonly();
        this.listen = listener => source.listen(listener);
        this.at = index => source.at(index)?.readonly();
        this.reverse = () => source.reverse();
        this.slice = (start, end) => source.slice(start, end);
        this.toArray = () => source.toArray();
        this.join = joiner => source.join(joiner);
    }
    *[Symbol.iterator]() {
        for (const member of this.source)
            yield member.readonly();
    }
    each(modifier) {
        const target = new List();
        for (const member of this)
            target.append(modifier(member));
        this.source.listen(update => {
            switch (update.kind) {
                case ListEventKind.Append:
                    return target.append(modifier(update.member.readonly()));
                case ListEventKind.Prepend:
                    return target.prepend(modifier(update.member.readonly()));
                case ListEventKind.Insert:
                    return target.insertAt(update.index, modifier(update.member.readonly()));
                case ListEventKind.Remove:
                    return target.removeAt(update.index);
                case ListEventKind.Remove:
                    return target.removeAt(update.index);
                case ListEventKind.Replace:
                    return target.replaceAt(update.index, modifier(update.member.readonly()));
                case ListEventKind.Swap:
                    return target.swapBetween(update.from, update.to);
                case ListEventKind.Move:
                    return target.moveTo(update.to, update.member);
                case ListEventKind.Update:
                    return target.set(update.list.array.map(x => modifier(x.readonly())));
                case ListEventKind.MemberUpdate:
                    return target.at(update.member.getCurrentIndex())?.set(modifier(update.member.readonly()));
                default:
                    unreachable(update);
            }
        });
        return target.readonly();
    }
    readonly() {
        return this;
    }
}
