export class ReadonlyState {
    source;
    subscribers = [];
    state;
    constructor(value, source) {
        this.source = source;
        this.state = { value };
        if (source)
            source.listen(value => (this.state.value = value));
    }
    get value() {
        return this.state.value;
    }
    to(transformer) {
        const s = new State(null);
        // publish transformed changes when value changes
        this.listen(value => s.set(transformer(value)));
        // This has to be done after the above listener is attached.
        // Otherwise if this.set is called within the transformer,
        // the listener will not be called, and `s` will not get updated
        s.set(transformer(this.value));
        // return readonly so transformed state can't be published into
        return s.readonly();
    }
    pipe(state) {
        this.listen(value => state.set(value));
    }
    static isState(x) {
        return x instanceof State || x instanceof ReadonlyState;
    }
    /**
     * Compose multiple states into a single state
     */
    static compose(refs) {
        // lazily initialised below
        const initialValue = {};
        const merged = new State(initialValue);
        for (const key in refs) {
            initialValue[key] = refs[key].value;
            refs[key].listen(updated => merged.setWith(value => ({ ...value, [key]: updated })));
        }
        return merged.readonly();
    }
    if(predicate) {
        const filtered = new State(predicate(this.value) ? this.value : null);
        this.listen(value => {
            if (predicate(value))
                filtered.set(value);
        });
        return filtered.readonly();
    }
    debounce(ms) {
        const debounced = new State(this.value);
        let timeout;
        this.listen(value => {
            clearTimeout(timeout);
            timeout = setTimeout(() => debounced.set(value), ms);
        });
        return debounced.readonly();
    }
    listen(listener) {
        if (this.source)
            return this.source.listen(listener);
        this.subscribers.push(listener);
        // Return cleanup function
        return () => {
            const index = this.subscribers.indexOf(listener);
            if (index > -1)
                this.subscribers.splice(index, 1);
        };
    }
}
export class State extends ReadonlyState {
    constructor(value) {
        super(value);
    }
    set(next) {
        if (this.state.value === next)
            return this.value;
        this.state.value = next;
        this.subscribers.forEach(subscriber => subscriber(this.value));
        return this.value;
    }
    setWith(updater) {
        return this.set(updater(this.value));
    }
    readonly() {
        return new ReadonlyState(this.value, this);
    }
}
